#==============================================================================
# Copyright 2011 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================
import os
import logging

def set_owner_and_group(filename, owner_name, group_name):
    logging.warn("Unsupported OS for setting owner/group: %s" % os.name)

def create_or_modify_user(user_name, groups=[], homedir=None, uid=None):
    logging.warn("Unsupported OS for user operations: %s", os.name)

def create_group(group_name, gid=None):
    logging.warn("Unsupported OS for group operations: %s", os.name)

if os.name == "posix":
    from posix_security import *
